/*
 * $Id$
 *
 * Copyright (C) 2009, 2010 Elias Norberg, Patrik Sternudd
 *
 * This file is part of TDEPS.
 *
 * TDEPS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TDEOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TDEPS.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gnutls/gnutls.h>
#include "comm.h"

int connection_tcp_connect(char *server, int port)
{

	int err, sd;
	struct sockaddr_in sa;

	sd = socket(AF_INET, SOCK_STREAM, 0);

	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	inet_pton(AF_INET, server, &sa.sin_addr);

	if( (err = connect(sd, (struct sockaddr*)&sa, sizeof(sa))) < 0 ){
		return -1;
	}

	return sd;
}

int connection_print_x509_cert_info(session)
{
	printf("X509 certificate information\n");
	return 0;
}

int connection_print_session_info(gnutls_session_t session)
{
	const char *tmp;
	gnutls_credentials_type_t cred_type;
	gnutls_kx_algorithm_t kx;

	kx = gnutls_kx_get(session);
	tmp = gnutls_kx_get_name(kx);

	printf("Key Exchange: %s\n", tmp);

	cred_type = gnutls_auth_get_type(session);

	switch(cred_type){
		case GNUTLS_CRD_IA:
			printf("TLS/IA session\n");
		break;
		case GNUTLS_CRD_ANON:
			printf("Anonymous DH using prime of %d bits\n",
				gnutls_dh_get_prime_bits(session));
		break;
		case GNUTLS_CRD_CERTIFICATE:
		if( kx == GNUTLS_KX_DHE_RSA || kx == GNUTLS_KX_DHE_DSS ){
			printf("Ephemeral DH using prime of %d bits \n",
				gnutls_dh_get_prime_bits(session));
		}
		connection_print_x509_cert_info(session);
		break;
		case GNUTLS_CRD_SRP:
		/* print SRP info */
		break;
		case GNUTLS_CRD_PSK:
		/* print PSK info */
		break;
	}

	tmp = gnutls_protocol_get_name(gnutls_protocol_get_version(session));
	printf("Protocol: %s\n", tmp);

	tmp = gnutls_certificate_type_get_name(gnutls_certificate_type_get(session));

	printf("Certificate type: %s\n", tmp);

	tmp = gnutls_compression_get_name(gnutls_compression_get(session));
	printf("Compression: %s\n", tmp);

	tmp = gnutls_cipher_get_name(gnutls_cipher_get(session));
	printf("Cipher: %s\n", tmp);

	tmp = gnutls_mac_get_name(gnutls_mac_get(session));
	printf("MAC: %s\n", tmp);

	return 0;
}

#define CAFILE "ca-cert.pem"
#define MSG "GET / HTTP/1.0\r\n\r\n\r\n"

/*!
 * connection_set_flag( conn_ptr connection, int flag, void *value);
 * 
 * Sets a flag on the connection
 * 
 * Returns:
 *  -1 on error
 */
int connection_set_flag( conn_ptr connection, int flag, void *value)
{
	if( flag > N_FLAGS || flag < 0 )
		return -1;

	connection->flags[flag] = value;

	return 0;
}

/*!
 * conn_ptr connection_create();
 *
 * Allocates memory for a connection and returns it.
 *
 * Returns:
 *  On success, a pointer to a newly allocated conn_t structure
 *  On failure, NULL
 */
conn_ptr connection_create()
{
	conn_ptr ptr;

	ptr = calloc(1, sizeof(conn_t));

	return ptr;
}

/*! 
 * connection_open( conn_ptr connection );
 * 
 * Opens a connection based on the flags previously set
 *
 * Returns:
 *  -1 on error
 */
int connection_open( conn_ptr connection )
{
	int ret, sd, ii;
	gnutls_session_t session;
	char buffer[1024];
	const char *err;
	gnutls_certificate_credentials_t xcred;

	gnutls_global_init();

	/* Check that the neccessary flags are set */
	if( connection->flags[C_FLAG_REMOTE_SERVER_ADDRESS] == NULL ||
	    connection->flags[C_FLAG_REMOTE_SERVER_PORT] == NULL ){
		fprintf(stderr,"connection_open(): Remote server address or port not set!\n");
		return -1;
	}

	if( connection->flags[C_FLAG_REMOTE_SERVER_CERT] == NULL ){
		fprintf(stderr,"connection_open(): Remote server certificate not set!\n");
		return -1;
	}

	if( (ret = gnutls_certificate_allocate_credentials(&xcred)) < 0 ){
		gnutls_perror(ret);
		return -1;
	}

	if( (ret = gnutls_certificate_set_x509_trust_file(xcred, CAFILE, GNUTLS_X509_FMT_PEM)) < 0 ){
		fprintf(stderr, "gnutls_certificate_set_x509_trust_file():");
		gnutls_perror(ret);
		return -1;
	}

	gnutls_init(&session, GNUTLS_CLIENT);

	ret = gnutls_priority_set_direct( session, "PERFORMANCE", &err );
	if( ret < 0 ){
		if( ret == GNUTLS_E_INVALID_REQUEST ){
			fprintf(stderr, "Syntax error at %s\n", err);
		}
		return 1;
	}

	gnutls_credentials_set( session, GNUTLS_CRD_CERTIFICATE, xcred );


	sd = connection_tcp_connect( (char *)connection->flags[C_FLAG_REMOTE_SERVER_ADDRESS], 
			*((int *)connection->flags[C_FLAG_REMOTE_SERVER_PORT]));
	if( sd == -1 ){
		perror("connection_tcp_connect()");
		return -1;
	}

	gnutls_transport_set_ptr ( session, (gnutls_transport_ptr_t) &sd );

	ret = gnutls_handshake( session );

	if( ret < 0 ){
		fprintf(stderr, "Handshake failed\n");
		gnutls_perror(ret);
		return -1;
	}
	printf("Handshake completed\n");

	connection_print_session_info(session);
	
	gnutls_record_send(session, "ABC", 3);

	ret = gnutls_record_recv(session, buffer, 1024);
	if( ret == 0 ){
		printf("Peer has closed the TLS connection\n");
		return -1;
	}else if( ret < 0 ){
		fprintf(stderr, "Error: %s\n", gnutls_strerror(ret));
		return -1;
	}

	printf("Recieved %d bytes: ", ret );
	for(ii=0;ii<ret;ii++)
		fputc(buffer[ii],stdout);
	fputc('\n',stdout);

	gnutls_bye(session, GNUTLS_SHUT_RDWR);

	close(sd);
	gnutls_deinit(session);
	gnutls_certificate_free_credentials(xcred);
	gnutls_global_deinit();

	return 0;
}
